;=====================================================================================;
;fpaq.inc
;constants and macros for fpaq.asm and fpaq0 compress/decompress algorithm.
;FASMW code implementation from Nikolay Petrov
;=====================================================================================;
;enum
COMPRESS	= 0
DECOMPRESS	= 1
virtual at esp
  x1	dd ?	       ;x1 = dword[esp]
  x2	dd ?	       ;x2 = dword[esp+4]
  ct	dd 256*2 dup(?);ct[256][2] - offset from (esp+8)
end virtual
;=====================================================================================;
macro case var,[const,lebel]
 {cmp		var,const
  je		lebel}
;=====================================================================================;
macro jneq var,[const,lebel]
 {cmp		var,const
  jne		lebel}
;=====================================================================================;
macro clock t_var
  {call        [GetTickCount]
   if ~ t_var eq
      sub	eax,t_var
   end if}
;=====================================================================================;
macro GetFpaqArgs
 {local ..search_firstpos,\
	..find_firstpos  ,\
	..search_endpos  ,\
	..find_all
  push		esi edi
  xor		edi,edi
  call		[GetCommandLine]
  mov		esi,eax
  movzx 	eax,byte[esi]
  xor		ecx,ecx
  xor		edx,edx
  jmp		..find_firstpos
..search_firstpos:
  test		eax,eax
  je		..find_all
  jneq		ax,' ',..find_firstpos
  mov		byte [esi],0
  inc		esi
  movzx 	eax,byte[esi]
  jmp		..search_firstpos
..find_firstpos:
  mov		dx,' '
  cmp		eax,'"'
  sete		cl
  add		esi,ecx
  lea		edx,[edx+2*ecx]
  mov		[argv+4*edi],esi; --> program name - dword [argv]
  inc		edi		;     first arg.   - dword [argv+4]
				;     second arg.  - dword [argv+8]...
  cmp		di,MAX_ARGUMENTS
  jae		PrintHelp
..search_endpos:
  lodsb
  test		eax,eax
  je		..find_all
  jneq		eax,edx,..search_endpos
  mov		byte [esi-1],0
  movzx 	eax,byte[esi]
  jmp		..search_firstpos
..find_all:
;edi = argc
  jneq		edi,MAX_ARGUMENTS-1,PrintHelp
  pop		edi esi}
;=====================================================================================;
;Constructor
macro InitEncoder _Mode
 {align   8
  .#_Mode:
  push		ebp ebx esi edi
  mov		[v_stack.save],esp ;save stack
  mov		esp,v_stack.esp ;set the new stack point
  ;x1=0;x=0;x2=-1
  mov		[x2],-1
  mov		esi,[infile.stream]
  mov		edi,outfile.stream
  if _Mode eq DECOMPRESS
     add	esi,4
     get_x
  end if
  .#_Mode#.loop:}
;=====================================================================================;
macro get_x
;ebx = x, esi - in
 {mov		ebx,[esi-4]
  bswap 	ebx}
;=====================================================================================;
;assert(xmid >= x1 && xmid < x2);
macro Assert var
 {if var eq x1
     jb 	assert
  else if var eq x2
     cmp	eax,[var]
     jae	assert
  end if}
;=====================================================================================;
macro predictor.p bit
;Predictor.p() = 4096*(ct[cxt][1]+1)/(ct[cxt][0]+ct[cxt][1]+2)
;return result in eax
 {if bit in <0,1>
     ;ecx = ct[0][0]
     inc	ecx
     if bit eq 0
	mov	[ct],ecx
     end if
     inc	ecx
     xor	eax,eax
     mov	ah,1 shl (P_SHIFT-8)
  else if bit eq
     ;ebp = cxt
     mov	eax,[ct+8*ebp+4]
     mov	ecx,[ct+8*ebp]
     inc	eax
     lea	ecx,[ecx+eax+1]
     shl	eax,P_SHIFT
  else
     err
  end if
  sub		edx,edx
  div		ecx}
;=====================================================================================;
macro calc_xmid bit
;xmid = x1 + ((x2-x1) >> 12)*p()
;do it assert check and return result in eax
 {predictor.p	bit
  if bit eq
     xor	ecx,ecx
  end if
  mov		ecx,[x2]
  sub		ecx,[x1]
  shr		ecx,P_SHIFT
  mul		ecx
  add		eax,[x1]
  Assert	x1
  Assert	x2}
;=====================================================================================;
; encode-decode extra bit
macro ed_extra_bit _Mode
;ebp = cxt = 1
;return - in eax - x2; in edx - x1
 {mov		eax,[x1]
  mov		ecx,[ct]
  mov		ebp,1
  cmp		ecx,(1 shl P_SHIFT)-1
  jb		@f
  Assert	x2
  if _Mode eq DECOMPRESS
     cmp	eax,ebx
     jae	.finish
  end if
  lea		edx,[eax+1]
  mov		eax,[x2]
  mov		[x1],edx
  inc		[ct]
  jneq		ecx,0fffeh,.#_Mode#.while
  shr		ecx,1
  mov		[ct],ecx
  jmp		.#_Mode#.while
align	 8
@@:
  calc_xmid	0
  if _Mode eq DECOMPRESS
     cmp	eax,ebx
     jae	.finish
  end if
  lea		edx,[eax+1]
  mov		[x1],edx
  mov		eax,[x2]
  jmp		.#_Mode#.while
align	8}
;=====================================================================================;
macro update _Mode,bit;Predictor.update
 {inc		[ct+8*ecx+4*bit]
  jneq		[ct+8*ecx+4*bit],0ffffh,.#_Mode#.while
  shr		[ct+8*ecx],1
  shr		[ct+8*ecx+4],1
  jmp		.#_Mode#.while
align	8}
